#!/usr/local/bin/python
# -*- Mode: Python; tab-width: 4 -*-

# most of the code comes from exifdump.py, v1.13:
#   Exif information decoder
#   Written by Thierry Bousch <bousch@topo.math.u-psud.fr>
#   Public Domain
#
# mangling by Sam Rushing <rushing@nightmare.com>
#

#
# fetch a thumbnail from an exif jpeg image. (jpeg format only)
#

def s2n_motorola(str):
	x = 0
	for c in str:
		x = (x << 8) | ord(c)
	return x

def s2n_intel(str):
	x = 0
	y = 0
	for c in str:
		x = x | (ord(c) << y)
		y = y + 8
	return x

class Fraction:

  def __init__(self, num, den):
    self.num = num
    self.den = den

  def __repr__(self):
    # String representation
    return '%d/%d' % (self.num, self.den)

class TIFF_file:

	def __init__(self, data):
		self.data = data
		self.endian = data[0]
	
	def s2n(self, offset, length, signed=0):
		slice = self.data[offset:offset+length]
		if self.endian == 'I':
			val = s2n_intel(slice)
		else:
			val = s2n_motorola(slice)
		# Sign extension ?
		if signed:
			msb = 1 << (8*length - 1)
			if val & msb:
				val = val - (msb << 1)
		return val

	def first_IFD(self):
		return self.s2n(4, 4)
	
	def next_IFD(self, ifd):
		entries = self.s2n(ifd, 2)
		return self.s2n(ifd + 2 + 12 * entries, 4)

	def list_IFDs(self):
		i = self.first_IFD()
		a = []
		while i:
			a.append(i)
			i = self.next_IFD(i)
		return a
		
	def dump_IFD(self, ifd):
		entries = self.s2n(ifd, 2)
		a = []
		for i in range(entries):
			entry = ifd + 2 + 12*i
			tag = self.s2n(entry, 2)
			type = self.s2n(entry+2, 2)
			if not 1 <= type <= 10:
				continue # not handled
			typelen = [ 1, 1, 2, 4, 8, 1, 1, 2, 4, 8 ] [type-1]
			count = self.s2n(entry+4, 4)
			offset = entry+8
			if count*typelen > 4:
				offset = self.s2n(offset, 4)
			if type == 2:
				# Special case: nul-terminated ASCII string
				values = self.data[offset:offset+count-1]
			else:
				values = []
				signed = (type == 6 or type >= 8)
				for j in range(count):
					if type % 5:
						# Not a fraction
						value_j = self.s2n(offset, typelen, signed)
					else:
						# The type is either 5 or 10
						value_j = Fraction(self.s2n(offset,	 4, signed),
															 self.s2n(offset+4, 4, signed))
					values.append(value_j)
					offset = offset + typelen
			# Now "values" is either a string or an array
			a.append((tag,type,values))
		return a

def get_marker (file):
	position = file.tell()
	marker = file.read (2)
	size = file.read (2)
	size = (ord(size[0])*256 + ord(size[1]))
	# size includes the two 'data-size' bytes
	return position, marker, file.read (size-2)

def fetch (file):
	head = file.read(2)
	if head != '\377\330':
		# FFD8 = SOI = start of image
		raise ValueError, 'Not a JPEG file'
	else:
		jfif = 0
		while 1:
			# search for Exif/APP1 marker (0xffe1)
			position, marker, data = get_marker (file)
			#print 'marker=%s length=%d' % (repr(marker), len(data))
			if marker == '\377\341':
				# skip over the next six bytes, should be "Exif\000\000"
				assert (data[:6] == "Exif\000\000")
				data = data[6:]
				break
			elif marker == '\377\340':
				jfif = 1
				
		# Exif uses TIFF
		T = TIFF_file (data)
		L = T.list_IFDs()
		# we want the second one.
		ifd = T.dump_IFD(L[1])
		offset = 0
		length = 0
		for type, tag, value in ifd:
			#0x201:	"JPEGInterchangeFormat",
			#0x202:	"JPEGInterchangeFormatLength",
			if type == 0x201:
				offset = value[0]
			elif type == 0x202:
				length = value[0]
		if (offset and length):
			return data[offset:offset+length]
		return None

if __name__ == '__main__':
	import sys
	sys.stdout.write (
		fetch (open (sys.argv[1], 'rb'))
		)
