# -*- Mode: Python; tab-width: 4 -*-

import coro
import coro_subproc
import coro_ehttpd
import os
import re
import string
import exifdump
import exif_thumb
import pprint_html
import session_handler

from urllib import quote_plus, unquote_plus

QP = quote_plus
UQP = unquote_plus

def split_path (path):
	# strip off leading slash, it's no fun!
	return string.split (path, '/')[1:]

class thumb_handler:

	def __init__ (self, doc_root):
		self.doc_root = doc_root
		self.thumb_cache = {}

	def match (self, request):
		path = split_path (UQP(request._path))
		if path and (path[0] == 'thumb'):
			request.path = path
			return 1
		else:
			return 0

	def get_thumb (self, filename):
		data = self.thumb_cache.get (filename, 'not-yet')
		# use an exif thumb if possible
		if data == 'not-yet':
			try:
				data = exif_thumb.fetch (open(filename, 'rb'))
			except:
				data = self.build_thumb (filename)
			self.thumb_cache[filename] = data
		return data

	def build_thumb (self, filename):
		sub = (
			'djpeg -fast -scale 1/8 \'%s\''
			' | cjpeg -quality 30 -progressive -dct fast' % (
				filename,
				)
			)
		fo, fi, fe, pid = coro_subproc.spawn_sub_process (sub)
		data = []
		while 1:
			block = fo.read (4096)
			if not block:
				os.waitpid (pid, os.WNOHANG)
				break
			else:
				data.append (block)
		return string.join (data, '')

	def handle_request (self, request):
		# try to fetch an exif thumbnail
		rpath = request.path
		spath = filter (None, rpath[1:])
		jspath = string.join (spath, '/')
		filename = apply (os.path.join, [self.doc_root] + spath)
		data = self.get_thumb (filename)
		if data:
			request['Content-Type'] = 'image/jpeg'
			request['Connection'] = 'close'
			request.push (data)
			request.done()
		else:
			request.error (404)

class file_handler:

	def __init__ (self, doc_root, sh):
		self.doc_root = doc_root
		self.sh = sh

	def match (self, request):
		path = split_path (UQP(request._path))
		if path and (path[0] == 'file'):
			request.path = path
			return 1
		else:
			return 0

	def handle_request (self, request):
		# just deliver the image file
		rpath = request.path
		spath = filter (None, rpath[1:])
		jspath = string.join (spath, '/')
		filename = apply (os.path.join, [self.doc_root] + spath)
		# defaults
		scale = 2
		grayscale = 0
		quality = 50

		if request._query:
			qd = frob_query (request._query)
			if qd.has_key ('q'):
				quality = int(qd['q'])
			if qd.has_key ('g'):
				grayscale = int(qd['g'])
			if qd.has_key ('s'):
				scale = int(qd['s'])

		if grayscale:
			grayscale = '-grayscale'
		else:
			grayscale = ''
		sub = (
			'djpeg %s -fast -scale 1/%d \'%s\''
			' | cjpeg -quality %d -progressive -dct fast' % (
				grayscale,
				scale,
				filename,
				quality
				)
			)
		fo, fi, fe, pid = coro_subproc.spawn_sub_process (sub)
		while 1:
			block = fo.read (4096)
			if not block:
				os.waitpid (pid, os.WNOHANG)
				break
			else:
				request._client.conn.send (block)
		request.done()

class zip_handler:

	def __init__ (self, doc_root):
		self.doc_root = doc_root

	def match (self, request):
		path = split_path (UQP(request._path))
		if path and (path[0] == 'zip'):
			request.path = path
			return 1
		else:
			return 0

	def handle_request (self, request):
		# try to fetch an exif thumbnail
		rpath = request.path
		spath = filter (None, rpath[1:])
		jspath = string.join (spath, '/')
		path = apply (os.path.join, [self.doc_root] + spath)
		if os.path.isdir (path):
			request['Content-Type'] = 'application/zip'
			request['Content-Disposition'] = 'inline; filename=%s.zip' % spath[-1]
			request['Connection'] = 'close'
			sub = ('zip -rj0 - %s/*.{jpeg,jpg,JPEG,JPG}' % path)
			fo, fi, fe, pid = coro_subproc.spawn_sub_process (sub)
			data = []
			while 1:
				block = fo.read (4096)
				if not block:
					os.waitpid (pid, os.WNOHANG)
					break
				else:
					request.push (block)
		else:
			request.error (404)
		request.done()

def is_jpeg (path):
	return string.split (path, '.')[-1] in ('jpg','JPG','jpeg', 'JPEG')

def frob_query (q):
	d = {}
	pairs = string.split (q, '&')
	for p in pairs:
		key, val = string.split (p, '=')
		d[key] = UQP(val)
	return d

selections = {
	'q':('quality',[(30,'low'),(50,'medium'),(80,'high')]),
	's':('scale',[(1,'1/1'),(2,'1/2'),(4,'1/4'),(8,'1/8')]),
	'g':('color',[(0,'color'),(1,'grayscale')]),
	}

def join_path (dir_list, tail=''):
	return string.join (dir_list + [tail], '/')

class image_session:

	def __init__ (self, doc_root, th):
		self.doc_root = doc_root
		self.th = th
		self.state = None

	def get_image_files (self, img_dir):
		img_path = os.path.normpath (apply (os.path.join, [self.doc_root] + img_dir))
		names = os.listdir (img_path)
		jpegs = []
		dirs = []
		for name in names:
			joined = os.path.join (img_path, name)
			if name[:5] != 'index' and is_jpeg (name):
				jpegs.append ((name, joined))
			elif os.path.isdir (joined):
				dirs.append (name)
		jpegs.sort()
		dirs.sort()
		return jpegs, dirs

	def session (self, handler, request):

		# global state
		scale = 4
		quality = 75
		grayscale = 1
		img_num = 0
		img_dir = []
		# are we looking at an image, or at an index?
		in_index = 1

		if self.state:
			# recover state
			grayscale, scale, quality, in_index, img_dir, img_num = self.state

		# initialize listing
		img_files, img_dirs  = self.get_image_files (img_dir)


		def bound (n, low, high):
			n = min (n, high)
			n = max (n, low)
			return n


		i = 0
		while 1:
			i = i + 1
			
			# ---------------------------------------------------------------------------
			#                           session variables
			# ---------------------------------------------------------------------------

			if request._query:
				qd = frob_query (request._query)
				if qd.has_key ('q'):
					quality = int(qd['q'])
				if qd.has_key ('g'):
					grayscale = int(qd['g'])
				if qd.has_key ('s'):
					scale = int(qd['s'])
				if qd.has_key ('p'):
					# multi-directory would be nice here
					path = qd['p']
					if path in img_dirs:
						# ignore bogus directories
						img_dir.append (path)
						img_files, img_dirs = self.get_image_files (img_dir)
						in_index = 1
						img_num = 0
				if qd.has_key ('m'):
					in_index = not int(qd['m'])
				if qd.has_key ('i'):
					img_num = bound (int(qd['i']), 0, len(img_files))
				if qd.has_key ('n'):
					img_num = bound (img_num + 1, 0, len(img_files))
				if qd.has_key ('p'):
					img_num = bound (img_num - 1, 0, len(img_files))
				if qd.has_key ('u'):
					img_dir = img_dir[:-1]
					img_files, img_dirs = self.get_image_files (img_dir)
					in_index = 1
					img_num = 0
				if qd.has_key ('r'):
					# refresh
					img_files, img_dirs = self.get_image_files (img_dir)
					
			request['Content-Type'] = 'text/html'
			request['Connection'] = 'close'

			request.push ('<html><h1>session %x.%d</h1>' % (id(self), i))
			request.push ('<h2>directory: %s</h2>\r\n' % ('/' + string.join (img_dir, '/')))

			if in_index:
				# ---------------------------------------------------------------------------
				#                                 index
				# ---------------------------------------------------------------------------
				if len(img_dir):
					request.push ('<a href="?u=1">parent directory</a> | <a href="?r=1">refresh</a>\r\n<br>\r\n')
				if img_dirs:
					ld = []
					for x in img_dirs:
						ld.append ('<a href="?p=%s">%s</a>' % (x, x))
					request.push (string.join (ld, ' | ') + '<hr>')
				for j in range(len(img_files)):
					name, path = img_files[j]
					data = th.get_thumb (path)
					if data:
						# build thumb path
						thumb_path = os.path.normpath (apply (os.path.join, img_dir + [name]))
						request.push ('<a href="?m=1&i=%d"><img src="/thumb/%s" alt="%s"></a>\r\n' % (j, QP(thumb_path), name))
					else:
						request.push ('<a href="?m=1&i=%d">%s</a> | \r\n' % (j, name))

				if len(img_files):
					# build zip path
					zip_path = join_path (img_dir)
					request.push ('<br><a href="/zip/%s">[zip this directory]</a>\r\n' % (QP(zip_path)))
			else:
				# ---------------------------------------------------------------------------
				#                                  image
				# ---------------------------------------------------------------------------
				name, path = img_files[img_num]
				prev = ((img_num-1) % len(img_files))
				next = ((img_num+1) % len(img_files))
				tpp = QP(os.path.normpath (apply (os.path.join, img_dir + [img_files[prev][0]])))
				tpn = QP(os.path.normpath (apply (os.path.join, img_dir + [img_files[next][0]])))
				request.push (
					'<ul><li><b>image:</b> %d<li><b>filename:</b> %s</ul>' % (img_num, name) +
					'<a href="?m=0">[back to index]</a>\r\n'
					)
				path = os.path.normpath (apply (os.path.join, img_dir + [name]))
				request.push (
					'<table><tr>'
					'<td>'
					'<a href="?i=%d">&lt;--[prev]<br><img src="/thumb/%s" alt="%s"></a><br>\r\n'
					'<a href="?i=%d"><img src="/thumb/%s" alt="%s"><br>[next]--&gt;</a></td>\r\n' % (prev, tpp, tpp, next, tpn, tpn) +
					'<td><img border=1 src="/file/%s?q=%d&s=%d&g=%d" alt="%s">\r\n' % (QP(path), quality, scale, grayscale, name) +
					'</td></table>'
					)

				l = ['<ul>']
				for k,(setting,v) in selections.items():
					l.append ('<li><b>%s</b>: ' % setting)
					l2 = []
					for val,name in v:
						l2.append (
							'<a href="?%s=%s">%s</a>' % (k, val, name)
							)
					l.append (string.join (l2, ' | '))
				l.append ('</ul>')
				request.push (string.join (l, '\r\n'))
				f = open (img_files[img_num][1], 'rb')
				f.seek (0,2)
				size = f.tell()
				request.push ('<b>file size:</b> %d bytes<br>' % (size))
				f.seek (0,0)
				try:
					d = exifdump.process_file (f)
				except:
					d = None
				f.close()
				if d:
					l = []
					pprint_html.pprint_html (d, l.append)
					request.push (string.join (l, ''))

			request.push ('</html>')
			request.done()
			# remember state
			self.state = (grayscale, scale, quality, in_index, img_dir, img_num)
			# fetch the next one
			request = handler.get_next_request()

if __name__ == '__main__':
	import marshal
	import time
	import sys

	def log_override (self, bytes):
		whom = '%s:%d ' % (self._whom)
		sys.stderr.write ('%s %s %d\n' % (whom, self._request, bytes))
		# kludge!
		return ''

	if len(sys.argv) != 3:
		print 'Usage: %s <doc_root> <port>' % sys.argv[0]
		sys.exit()
	else:
		doc_root = sys.argv[1]
		port = int(sys.argv[2])

	coro_ehttpd.http_request.log = log_override
	server = coro_ehttpd.http_server (args = (('', port),))

	th = thumb_handler (doc_root)
	saved_state = {}

	def make_new_session (handler, request, sid):
		s = image_session (doc_root, th)
		me = coro.current_thread()
		me.session = s
		if saved_state.has_key (sid):
			s.state = saved_state[sid]

		handler.sessions[sid] = me
		s.session (handler, request)
		return s

	root = image_session (doc_root, th)
	sh = session_handler.session_handler ('image', make_new_session)
	fh = file_handler (doc_root, sh)
	zh = zip_handler (doc_root)
	server.push_handler (th)
	server.push_handler (fh)
	server.push_handler (zh)
	server.push_handler (sh)
	
	server.start()
	try:
		f = open (os.path.join (doc_root, '.http_state.%d.bin' % port), 'rb')
		saved = marshal.load (f)
		f.close()
		for (sid, state, when) in saved:
			saved_state[sid] = state
	except:
		pass

	# Go!
	try:
		coro.event_loop (30.0)
	finally:
		# save state
		saved = []
		sessions = sh.sessions.items()
		now = int(time.time())
		for i in range(len(sessions)):
			sid, c = sessions[i]
			saved.append ((sid, c.session.state, now))
		f = open (os.path.join (doc_root, '.http_state.bin'), 'wb')
		marshal.dump (saved, f)
		f.close()
